@@ -11,30 +11,20 @@ class VerificationMissing(Exception):
     """
 
 
-def get_extension(srcfilename, modname, **kwds):
+def get_extension(srcfilename, modname, sources=(), **kwds):
     from distutils.core import Extension
-    return Extension(name=modname, sources=[srcfilename], **kwds)
+    allsources = [srcfilename]
+    allsources.extend(sources)
+    return Extension(name=modname, sources=allsources, **kwds)
 
 def compile(tmpdir, ext):
     """Compile a C extension module using distutils."""
 
-    # Turn the 'ext.sources' into absolute paths, because we're going to
-    # do chdir().  In the common case where the path is precisely where
-    # we're going to chdir(), then replace it with a pathless copy.
-    for i, src in enumerate(ext.sources):
-        src = os.path.abspath(src)
-        if samefile(os.path.dirname(src), tmpdir):
-            src = os.path.basename(src)
-        ext.sources[i] = src
-
     saved_environ = os.environ.copy()
-    saved_path = os.getcwd()
     try:
-        os.chdir(tmpdir)
-        outputfilename = _build(ext)
+        outputfilename = _build(tmpdir, ext)
         outputfilename = os.path.abspath(outputfilename)
     finally:
-        os.chdir(saved_path)
         # workaround for a distutils bugs where some env vars can
         # become longer and longer every time it is used
         for key, value in saved_environ.items():
@@ -42,7 +32,7 @@ def compile(tmpdir, ext):
                 os.environ[key] = value
     return outputfilename
 
-def _build(ext):
+def _build(tmpdir, ext):
     # XXX compact but horrible :-(
     from distutils.core import Distribution
     import distutils.errors
@@ -50,6 +40,8 @@ def _build(ext):
     dist = Distribution({'ext_modules': [ext]})
     options = dist.get_option_dict('build_ext')
     options['force'] = ('ffiplatform', True)
+    options['build_lib'] = ('ffiplatform', tmpdir)
+    options['build_temp'] = ('ffiplatform', tmpdir)
     #
     try:
         dist.run_command('build_ext')
@@ -66,3 +58,21 @@ try:
 except ImportError:
     def samefile(f1, f2):
         return os.path.abspath(f1) == os.path.abspath(f2)
+
+def maybe_relative_path(path):
+    if not os.path.isabs(path):
+        return path      # already relative
+    dir = path
+    names = []
+    while True:
+        prevdir = dir
+        dir, name = os.path.split(prevdir)
+        if dir == prevdir or not dir:
+            return path     # failed to make it relative
+        names.append(name)
+        try:
+            if samefile(dir, os.curdir):
+                names.reverse()
+                return os.path.join(*names)
+        except OSError:
+            pass
