@@ -10,6 +10,7 @@ class VGenericEngine(object):
         self.verifier = verifier
         self.ffi = verifier.ffi
         self.export_symbols = []
+        self._struct_pending_verification = {}
 
     def patch_extension_kwds(self, kwds):
         # add 'export_symbols' to the dictionary.  Note that we add the
@@ -52,6 +53,7 @@ class VGenericEngine(object):
         # build the FFILibrary class and instance
         class FFILibrary(object):
             _cffi_generic_module = module
+            _cffi_ffi = self.ffi
         library = FFILibrary()
         #
         # finally, call the loaded_gen_xxx() functions.  This will set
@@ -108,13 +110,15 @@ class VGenericEngine(object):
             if isinstance(type, model.StructOrUnion):
                 indirection = '*'
             argnames.append('%sx%d' % (indirection, i))
-        arglist = [type.get_c_name(' %s' % arg)
+        context = 'argument of %s' % name
+        arglist = [type.get_c_name(' %s' % arg, context)
                    for type, arg in zip(tp.args, argnames)]
         arglist = ', '.join(arglist) or 'void'
         wrappername = '_cffi_f_%s' % name
         self.export_symbols.append(wrappername)
         funcdecl = ' %s(%s)' % (wrappername, arglist)
-        prnt(tp.result.get_c_name(funcdecl))
+        context = 'result of %s' % name
+        prnt(tp.result.get_c_name(funcdecl, context))
         prnt('{')
         #
         if not isinstance(tp.result, model.VoidType):
@@ -170,6 +174,16 @@ class VGenericEngine(object):
     def _loaded_gen_struct(self, tp, name, module, **kwds):
         self._loaded_struct_or_union(tp)
 
+    def _generate_gen_union_decl(self, tp, name):
+        assert name == tp.name
+        self._generate_struct_or_union_decl(tp, 'union', name)
+
+    def _loading_gen_union(self, tp, name, module):
+        self._loading_struct_or_union(tp, 'union', name, module)
+
+    def _loaded_gen_union(self, tp, name, module, **kwds):
+        self._loaded_struct_or_union(tp)
+
     def _generate_struct_or_union_decl(self, tp, prefix, name):
         if tp.fldnames is None:
             return     # nothing to do with opaque structs
@@ -181,9 +195,7 @@ class VGenericEngine(object):
         prnt('static void %s(%s *p)' % (checkfuncname, cname))
         prnt('{')
         prnt('  /* only to generate compile-time warnings or errors */')
-        for i in range(len(tp.fldnames)):
-            fname = tp.fldnames[i]
-            ftype = tp.fldtypes[i]
+        for fname, ftype, _ in tp.enumfields():
             if (isinstance(ftype, model.PrimitiveType)
                 and ftype.is_integer_type()):
                 # accept all integers, but complain on float or double
@@ -192,46 +204,27 @@ class VGenericEngine(object):
                 # only accept exactly the type declared.  Note the parentheses
                 # around the '*tmp' below.  In most cases they are not needed
                 # but don't hurt --- except test_struct_array_field.
-                prnt('  { %s = &p->%s; (void)tmp; }' % (
-                    ftype.get_c_name('(*tmp)'), fname))
+                try:
+                    prnt('  { %s = &p->%s; (void)tmp; }' % (
+                        ftype.get_c_name('(*tmp)', 'field %r'%fname), fname))
+                except ffiplatform.VerificationError as e:
+                    prnt('  /* %s */' % str(e))   # cannot verify it, ignore
         prnt('}')
         self.export_symbols.append(layoutfuncname)
         prnt('ssize_t %s(ssize_t i)' % (layoutfuncname,))
         prnt('{')
         prnt('  struct _cffi_aligncheck { char x; %s y; };' % cname)
-        if tp.partial:
-            prnt('  static ssize_t nums[] = {')
-            prnt('    1, sizeof(%s),' % cname)
-            prnt('    offsetof(struct _cffi_aligncheck, y),')
-            for fname in tp.fldnames:
-                prnt('    offsetof(%s, %s),' % (cname, fname))
-                prnt('    sizeof(((%s *)0)->%s),' % (cname, fname))
-            prnt('    -1')
-            prnt('  };')
-            prnt('  return nums[i];')
-        else:
-            ffi = self.ffi
-            BStruct = ffi._get_cached_btype(tp)
-            conditions = [
-                'sizeof(%s) != %d' % (cname, ffi.sizeof(BStruct)),
-                'offsetof(struct _cffi_aligncheck, y) != %d' % (
-                    ffi.alignof(BStruct),)]
-            for fname, ftype in zip(tp.fldnames, tp.fldtypes):
-                BField = ffi._get_cached_btype(ftype)
-                conditions += [
-                    'offsetof(%s, %s) != %d' % (
-                        cname, fname, ffi.offsetof(BStruct, fname)),
-                    'sizeof(((%s *)0)->%s) != %d' % (
-                        cname, fname, ffi.sizeof(BField))]
-            prnt('  if (%s ||' % conditions[0])
-            for i in range(1, len(conditions)-1):
-                prnt('      %s ||' % conditions[i])
-            prnt('      %s) {' % conditions[-1])
-            prnt('    return -1;')
-            prnt('  }')
-            prnt('  else {')
-            prnt('    return 0;')
-            prnt('  }')
+        prnt('  static ssize_t nums[] = {')
+        prnt('    sizeof(%s),' % cname)
+        prnt('    offsetof(struct _cffi_aligncheck, y),')
+        for fname, _, fbitsize in tp.enumfields():
+            if fbitsize >= 0:
+                continue      # xxx ignore fbitsize for now
+            prnt('    offsetof(%s, %s),' % (cname, fname))
+            prnt('    sizeof(((%s *)0)->%s),' % (cname, fname))
+        prnt('    -1')
+        prnt('  };')
+        prnt('  return nums[i];')
         prnt('  /* the next line is not executed, but compiled */')
         prnt('  %s(0);' % (checkfuncname,))
         prnt('}')
@@ -241,48 +234,80 @@ class VGenericEngine(object):
         if tp.fldnames is None:
             return     # nothing to do with opaque structs
         layoutfuncname = '_cffi_layout_%s_%s' % (prefix, name)
-        cname = ('%s %s' % (prefix, name)).strip()
         #
         BFunc = self.ffi.typeof("ssize_t(*)(ssize_t)")
         function = module.load_function(BFunc, layoutfuncname)
-        layout = function(0)
-        if layout < 0:
-            raise ffiplatform.VerificationError(
-                "incompatible layout for %s" % cname)
-        elif layout == 0:
-            assert not tp.partial
-        else:
-            totalsize = function(1)
-            totalalignment = function(2)
-            fieldofs = []
-            fieldsize = []
-            num = 3
-            while True:
-                x = function(num)
-                if x < 0: break
-                fieldofs.append(x)
-                fieldsize.append(function(num+1))
-                num += 2
+        layout = []
+        num = 0
+        while True:
+            x = function(num)
+            if x < 0: break
+            layout.append(x)
+            num += 1
+        if isinstance(tp, model.StructType) and tp.partial:
+            # use the function()'s sizes and offsets to guide the
+            # layout of the struct
+            totalsize = layout[0]
+            totalalignment = layout[1]
+            fieldofs = layout[2::2]
+            fieldsize = layout[3::2]
+            tp.force_flatten()
             assert len(fieldofs) == len(fieldsize) == len(tp.fldnames)
             tp.fixedlayout = fieldofs, fieldsize, totalsize, totalalignment
+        else:
+            cname = ('%s %s' % (prefix, name)).strip()
+            self._struct_pending_verification[tp] = layout, cname
 
     def _loaded_struct_or_union(self, tp):
         if tp.fldnames is None:
             return     # nothing to do with opaque structs
         self.ffi._get_cached_btype(tp)   # force 'fixedlayout' to be considered
 
+        if tp in self._struct_pending_verification:
+            # check that the layout sizes and offsets match the real ones
+            def check(realvalue, expectedvalue, msg):
+                if realvalue != expectedvalue:
+                    raise ffiplatform.VerificationError(
+                        "in %s: %s (we have %d, but C compiler says %d)"
+                        % (cname, msg, expectedvalue, realvalue))
+            ffi = self.ffi
+            BStruct = ffi._get_cached_btype(tp)
+            layout, cname = self._struct_pending_verification.pop(tp)
+            check(layout[0], ffi.sizeof(BStruct), "wrong total size")
+            check(layout[1], ffi.alignof(BStruct), "wrong total alignment")
+            i = 2
+            for fname, ftype, fbitsize in tp.enumfields():
+                if fbitsize >= 0:
+                    continue        # xxx ignore fbitsize for now
+                check(layout[i], ffi.offsetof(BStruct, fname),
+                      "wrong offset for field %r" % (fname,))
+                BField = ffi._get_cached_btype(ftype)
+                check(layout[i+1], ffi.sizeof(BField),
+                      "wrong size for field %r" % (fname,))
+                i += 2
+            assert i == len(layout)
+
     # ----------
     # 'anonymous' declarations.  These are produced for anonymous structs
     # or unions; the 'name' is obtained by a typedef.
 
     def _generate_gen_anonymous_decl(self, tp, name):
-        self._generate_struct_or_union_decl(tp, '', name)
+        if isinstance(tp, model.EnumType):
+            self._generate_gen_enum_decl(tp, name, '')
+        else:
+            self._generate_struct_or_union_decl(tp, '', name)
 
     def _loading_gen_anonymous(self, tp, name, module):
-        self._loading_struct_or_union(tp, '', name, module)
+        if isinstance(tp, model.EnumType):
+            self._loading_gen_enum(tp, name, module, '')
+        else:
+            self._loading_struct_or_union(tp, '', name, module)
 
     def _loaded_gen_anonymous(self, tp, name, module, **kwds):
-        self._loaded_struct_or_union(tp)
+        if isinstance(tp, model.EnumType):
+            self._loaded_gen_enum(tp, name, module, **kwds)
+        else:
+            self._loaded_struct_or_union(tp)
 
     # ----------
     # constants, likely declared with '#define'
@@ -300,7 +325,7 @@ class VGenericEngine(object):
             prnt('}')
         else:
             assert tp is not None
-            prnt(tp.get_c_name(' %s(void)' % funcname),)
+            prnt(tp.get_c_name(' %s(void)' % funcname, name),)
             prnt('{')
             if category == 'var':
                 ampersand = '&'
@@ -327,7 +352,7 @@ class VGenericEngine(object):
             if value < 0 and not negative:
                 value += (1 << (8*self.ffi.sizeof("long long")))
         else:
-            BFunc = self.ffi.typeof(tp.get_c_name('(*)(void)'))
+            BFunc = self.ffi.typeof(tp.get_c_name('(*)(void)', name))
             function = module.load_function(BFunc, funcname)
             value = function()
         return value
@@ -340,13 +365,13 @@ class VGenericEngine(object):
     # ----------
     # enums
 
-    def _generate_gen_enum_decl(self, tp, name):
+    def _generate_gen_enum_decl(self, tp, name, prefix='enum'):
         if tp.partial:
             for enumerator in tp.enumerators:
                 self._generate_gen_const(True, enumerator)
             return
         #
-        funcname = '_cffi_enum_%s' % name
+        funcname = '_cffi_e_%s_%s' % (prefix, name)
         self.export_symbols.append(funcname)
         prnt = self._prnt
         prnt('int %s(char *out_error)' % funcname)
@@ -363,17 +388,15 @@ class VGenericEngine(object):
         prnt('}')
         prnt()
 
-    _loading_gen_enum = _loaded_noop
-
-    def _loading_gen_enum(self, tp, name, module):
+    def _loading_gen_enum(self, tp, name, module, prefix='enum'):
         if tp.partial:
             enumvalues = [self._load_constant(True, tp, enumerator, module)
                           for enumerator in tp.enumerators]
             tp.enumvalues = tuple(enumvalues)
-            tp.partial = False
+            tp.partial_resolved = True
         else:
             BFunc = self.ffi.typeof("int(*)(char*)")
-            funcname = '_cffi_enum_%s' % name
+            funcname = '_cffi_e_%s_%s' % (prefix, name)
             function = module.load_function(BFunc, funcname)
             p = self.ffi.new("char[]", 256)
             if function(p) < 0:
@@ -427,7 +450,7 @@ class VGenericEngine(object):
         # remove ptr=<cdata 'int *'> from the library instance, and replace
         # it by a property on the class, which reads/writes into ptr[0].
         funcname = '_cffi_var_%s' % name
-        BFunc = self.ffi.typeof(tp.get_c_name('*(*)(void)'))
+        BFunc = self.ffi.typeof(tp.get_c_name('*(*)(void)', name))
         function = module.load_function(BFunc, funcname)
         ptr = function()
         def getter(library):
@@ -455,6 +478,7 @@ typedef unsigned __int16 uint16_t;
 typedef unsigned __int32 uint32_t;
 typedef unsigned __int64 uint64_t;
 typedef SSIZE_T ssize_t;
+typedef unsigned char _Bool;
 #else
 #  include <stdint.h>
 #endif
