@@ -1,6 +1,7 @@
 import py
 import sys, ctypes
 from cffi import FFI, CDefError
+from testing.support import *
 
 SIZE_OF_INT   = ctypes.sizeof(ctypes.c_int)
 SIZE_OF_LONG  = ctypes.sizeof(ctypes.c_long)
@@ -8,10 +9,6 @@ SIZE_OF_SHORT = ctypes.sizeof(ctypes.c_short)
 SIZE_OF_PTR   = ctypes.sizeof(ctypes.c_void_p)
 SIZE_OF_WCHAR = ctypes.sizeof(ctypes.c_wchar)
 
-if sys.version_info >= (3,):
-    unicode = str
-    long = int
-
 
 class BackendTests:
 
@@ -285,7 +282,7 @@ class BackendTests:
         assert int(ffi.cast("char", 300)) == 300 - 256
         assert bool(ffi.cast("char", 0))
         py.test.raises(TypeError, ffi.new, "char*", 32)
-        py.test.raises(TypeError, ffi.new, "char*", u"x")
+        py.test.raises(TypeError, ffi.new, "char*", u+"x")
         py.test.raises(TypeError, ffi.new, "char*", b"foo")
         #
         p = ffi.new("char[]", [b'a', b'b', b'\x9c'])
@@ -316,53 +313,53 @@ class BackendTests:
     def test_wchar_t(self):
         ffi = FFI(backend=self.Backend())
         self.check_wchar_t(ffi)
-        assert ffi.new("wchar_t*", u'x')[0] == u'x'
-        assert ffi.new("wchar_t*", u'\u1234')[0] == u'\u1234'
+        assert ffi.new("wchar_t*", u+'x')[0] == u+'x'
+        assert ffi.new("wchar_t*", u+'\u1234')[0] == u+'\u1234'
         if SIZE_OF_WCHAR > 2:
-            assert ffi.new("wchar_t*", u'\U00012345')[0] == u'\U00012345'
+            assert ffi.new("wchar_t*", u+'\U00012345')[0] == u+'\U00012345'
         else:
-            py.test.raises(TypeError, ffi.new, "wchar_t*", u'\U00012345')
-        assert ffi.new("wchar_t*")[0] == u'\x00'
+            py.test.raises(TypeError, ffi.new, "wchar_t*", u+'\U00012345')
+        assert ffi.new("wchar_t*")[0] == u+'\x00'
         assert int(ffi.cast("wchar_t", 300)) == 300
         assert bool(ffi.cast("wchar_t", 0))
         py.test.raises(TypeError, ffi.new, "wchar_t*", 32)
         py.test.raises(TypeError, ffi.new, "wchar_t*", "foo")
         #
-        p = ffi.new("wchar_t[]", [u'a', u'b', u'\u1234'])
+        p = ffi.new("wchar_t[]", [u+'a', u+'b', u+'\u1234'])
         assert len(p) == 3
-        assert p[0] == u'a'
-        assert p[1] == u'b' and type(p[1]) is type(u'')
-        assert p[2] == u'\u1234'
-        p[0] = u'x'
-        assert p[0] == u'x' and type(p[0]) is type(u'')
-        p[1] = u'\u1357'
-        assert p[1] == u'\u1357'
-        p = ffi.new("wchar_t[]", u"abcd")
+        assert p[0] == u+'a'
+        assert p[1] == u+'b' and type(p[1]) is unicode
+        assert p[2] == u+'\u1234'
+        p[0] = u+'x'
+        assert p[0] == u+'x' and type(p[0]) is unicode
+        p[1] = u+'\u1357'
+        assert p[1] == u+'\u1357'
+        p = ffi.new("wchar_t[]", u+"abcd")
         assert len(p) == 5
-        assert p[4] == u'\x00'
-        p = ffi.new("wchar_t[]", u"a\u1234b")
+        assert p[4] == u+'\x00'
+        p = ffi.new("wchar_t[]", u+"a\u1234b")
         assert len(p) == 4
-        assert p[1] == u'\u1234'
+        assert p[1] == u+'\u1234'
         #
-        p = ffi.new("wchar_t[]", u'\U00023456')
+        p = ffi.new("wchar_t[]", u+'\U00023456')
         if SIZE_OF_WCHAR == 2:
             assert sys.maxunicode == 0xffff
             assert len(p) == 3
-            assert p[0] == u'\ud84d'
-            assert p[1] == u'\udc56'
-            assert p[2] == u'\x00'
+            assert p[0] == u+'\ud84d'
+            assert p[1] == u+'\udc56'
+            assert p[2] == u+'\x00'
         else:
             assert len(p) == 2
-            assert p[0] == u'\U00023456'
-            assert p[1] == u'\x00'
+            assert p[0] == u+'\U00023456'
+            assert p[1] == u+'\x00'
         #
-        p = ffi.new("wchar_t[4]", u"ab")
+        p = ffi.new("wchar_t[4]", u+"ab")
         assert len(p) == 4
-        assert [p[i] for i in range(4)] == [u'a', u'b', u'\x00', u'\x00']
-        p = ffi.new("wchar_t[2]", u"ab")
+        assert [p[i] for i in range(4)] == [u+'a', u+'b', u+'\x00', u+'\x00']
+        p = ffi.new("wchar_t[2]", u+"ab")
         assert len(p) == 2
-        assert [p[i] for i in range(2)] == [u'a', u'b']
-        py.test.raises(IndexError, ffi.new, "wchar_t[2]", u"abc")
+        assert [p[i] for i in range(2)] == [u+'a', u+'b']
+        py.test.raises(IndexError, ffi.new, "wchar_t[2]", u+"abc")
 
     def test_none_as_null_doesnt_work(self):
         ffi = FFI(backend=self.Backend())
@@ -565,10 +562,10 @@ class BackendTests:
     def test_unicode_from_wchar_pointer(self):
         ffi = FFI(backend=self.Backend())
         self.check_wchar_t(ffi)
-        x = ffi.new("wchar_t*", u"x")
+        x = ffi.new("wchar_t*", u+"x")
         assert unicode(x) == unicode(repr(x))
-        assert ffi.string(x) == u"x"
-        assert ffi.string(ffi.new("wchar_t*", u"\x00")) == u""
+        assert ffi.string(x) == u+"x"
+        assert ffi.string(ffi.new("wchar_t*", u+"\x00")) == u+""
 
     def test_string_from_char_array(self):
         ffi = FFI(backend=self.Backend())
@@ -590,29 +587,29 @@ class BackendTests:
     def test_string_from_wchar_array(self):
         ffi = FFI(backend=self.Backend())
         self.check_wchar_t(ffi)
-        assert ffi.string(ffi.cast("wchar_t", "x")) == u"x"
-        assert ffi.string(ffi.cast("wchar_t", u"x")) == u"x"
+        assert ffi.string(ffi.cast("wchar_t", "x")) == u+"x"
+        assert ffi.string(ffi.cast("wchar_t", u+"x")) == u+"x"
         x = ffi.cast("wchar_t", "x")
         assert str(x) == repr(x)
-        assert ffi.string(x) == u"x"
+        assert ffi.string(x) == u+"x"
         #
-        p = ffi.new("wchar_t[]", u"hello.")
-        p[5] = u'!'
-        assert ffi.string(p) == u"hello!"
-        p[6] = u'\u04d2'
-        assert ffi.string(p) == u"hello!\u04d2"
-        p[3] = u'\x00'
-        assert ffi.string(p) == u"hel"
-        assert ffi.string(p, 123) == u"hel"
-        py.test.raises(IndexError, "p[7] = u'X'")
+        p = ffi.new("wchar_t[]", u+"hello.")
+        p[5] = u+'!'
+        assert ffi.string(p) == u+"hello!"
+        p[6] = u+'\u04d2'
+        assert ffi.string(p) == u+"hello!\u04d2"
+        p[3] = u+'\x00'
+        assert ffi.string(p) == u+"hel"
+        assert ffi.string(p, 123) == u+"hel"
+        py.test.raises(IndexError, "p[7] = u+'X'")
         #
-        a = ffi.new("wchar_t[]", u"hello\x00world")
+        a = ffi.new("wchar_t[]", u+"hello\x00world")
         assert len(a) == 12
         p = ffi.cast("wchar_t *", a)
-        assert ffi.string(p) == u'hello'
-        assert ffi.string(p, 123) == u'hello'
-        assert ffi.string(p, 5) == u'hello'
-        assert ffi.string(p, 2) == u'he'
+        assert ffi.string(p) == u+'hello'
+        assert ffi.string(p, 123) == u+'hello'
+        assert ffi.string(p, 5) == u+'hello'
+        assert ffi.string(p, 2) == u+'he'
 
     def test_fetch_const_char_p_field(self):
         # 'const' is ignored so far
@@ -631,10 +628,10 @@ class BackendTests:
         ffi = FFI(backend=self.Backend())
         self.check_wchar_t(ffi)
         ffi.cdef("struct foo { const wchar_t *name; };")
-        t = ffi.new("const wchar_t[]", u"testing")
+        t = ffi.new("const wchar_t[]", u+"testing")
         s = ffi.new("struct foo*", [t])
         assert type(s.name) not in (bytes, str, unicode)
-        assert ffi.string(s.name) == u"testing"
+        assert ffi.string(s.name) == u+"testing"
         s.name = ffi.NULL
         assert s.name == ffi.NULL
 
@@ -661,7 +658,6 @@ class BackendTests:
 
     def test_functionptr_simple(self):
         ffi = FFI(backend=self.Backend())
-        py.test.raises(TypeError, ffi.callback, "int(*)(int)")
         py.test.raises(TypeError, ffi.callback, "int(*)(int)", 0)
         def cb(n):
             return n + 1
@@ -751,14 +747,14 @@ class BackendTests:
     def test_wchar_cast(self):
         ffi = FFI(backend=self.Backend())
         self.check_wchar_t(ffi)
-        p = ffi.cast("int", ffi.cast("wchar_t", u'\u1234'))
+        p = ffi.cast("int", ffi.cast("wchar_t", u+'\u1234'))
         assert int(p) == 0x1234
         p = ffi.cast("long long", ffi.cast("wchar_t", -1))
         if SIZE_OF_WCHAR == 2:      # 2 bytes, unsigned
             assert int(p) == 0xffff
         else:                       # 4 bytes, signed
             assert int(p) == -1
-        p = ffi.cast("int", u'\u1234')
+        p = ffi.cast("int", u+'\u1234')
         assert int(p) == 0x1234
 
     def test_cast_array_to_charp(self):
@@ -878,6 +874,9 @@ class BackendTests:
         assert ffi.cast("enum bar", "A") != ffi.cast("int", 0)
         assert repr(ffi.cast("enum bar", "CC")) == "<cdata 'enum bar' 'CC'>"
         py.test.raises(ValueError, ffi.cast, "enum bar", "UNKNOWN")
+        ffi.cdef("enum baz { A=0x1000, B=0x2000 };")
+        assert int(ffi.cast("enum baz", "A")) == 0x1000
+        assert int(ffi.cast("enum baz", "B")) == 0x2000
 
     def test_enum_in_struct(self):
         ffi = FFI(backend=self.Backend())
@@ -1070,19 +1069,16 @@ class BackendTests:
             b = ffi.buffer(a)
         except NotImplementedError as e:
             py.test.skip(str(e))
-        if sys.version < '3':
-            assert type(b) is buffer
-            content = str(b)
-        else:
-            assert type(b) is memoryview
-            content = b.tobytes()
-        assert len(content) == 2
+        content = b[:]
+        assert len(content) == len(b) == 2
         if sys.byteorder == 'little':
             assert content == b'\x64\x00'
-            b[0] = b'\x65'[0]
+            assert b[0] == b'\x64'
+            b[0] = b'\x65'
         else:
             assert content == b'\x00\x64'
-            b[1] = b'\x65'[0]
+            assert b[1] == b'\x64'
+            b[1] = b'\x65'
         assert a[0] == 101
 
     def test_ffi_buffer_array(self):
@@ -1092,18 +1088,13 @@ class BackendTests:
             b = ffi.buffer(a)
         except NotImplementedError as e:
             py.test.skip(str(e))
-        if sys.version < '3':
-            assert type(b) is buffer
-            content = str(b)
-        else:
-            assert type(b) is memoryview
-            content = b.tobytes()
+        content = b[:]
         if sys.byteorder == 'little':
             assert content.startswith(b'\x64\x00\x00\x00\x65\x00\x00\x00')
-            b[4] = b'\x45'[0]
+            b[4] = b'\x45'
         else:
             assert content.startswith('\x00\x00\x00\x64\x00\x00\x00\x65')
-            b[7] = b'\x45'[0]
+            b[7] = b'\x45'
         assert len(content) == 4 * 10
         assert a[1] == 0x45
 
@@ -1114,20 +1105,15 @@ class BackendTests:
             b = ffi.buffer(a, 1)
         except NotImplementedError as e:
             py.test.skip(str(e))
-        if sys.version < '3':
-            assert type(b) is buffer
-            content = str(b)
-        else:
-            assert type(b) is memoryview
-            content = b.tobytes()
+        content = b[:]
         assert len(content) == 1
         if sys.byteorder == 'little':
             assert content == b'\x43'
-            b[0] = b'\x62'[0]
+            b[0] = b'\x62'
             assert a[0] == 0x4262
         else:
             assert content == b'\x42'
-            b[0] = b'\x63'[0]
+            b[0] = b'\x63'
             assert a[0] == 0x6343
 
     def test_ffi_buffer_array_size(self):
@@ -1138,10 +1124,7 @@ class BackendTests:
             ffi.buffer(a1)
         except NotImplementedError as e:
             py.test.skip(str(e))
-        if sys.version < '3':
-            assert str(ffi.buffer(a1)) == str(ffi.buffer(a2, 4*10))
-        else:
-            assert ffi.buffer(a1).tobytes() == ffi.buffer(a2, 4*10).tobytes()
+        assert ffi.buffer(a1)[:] == ffi.buffer(a2, 4*10)[:]
 
     def test_ffi_buffer_with_file(self):
         ffi = FFI(backend=self.Backend())
@@ -1163,6 +1146,24 @@ class BackendTests:
         f.close()
         os.unlink(filename)
 
+    def test_ffi_buffer_with_io(self):
+        ffi = FFI(backend=self.Backend())
+        import io, array
+        f = io.BytesIO()
+        a = ffi.new("int[]", list(range(1005)))
+        try:
+            ffi.buffer(a, 512)
+        except NotImplementedError as e:
+            py.test.skip(str(e))
+        f.write(ffi.buffer(a, 1000 * ffi.sizeof("int")))
+        f.seek(0)
+        assert f.read() == array.array('i', range(1000)).tostring()
+        f.seek(0)
+        b = ffi.new("int[]", 1005)
+        f.readinto(ffi.buffer(b, 1000 * ffi.sizeof("int")))
+        assert list(a)[:1000] + [0] * (len(a)-1000) == list(b)
+        f.close()
+
     def test_array_in_struct(self):
         ffi = FFI(backend=self.Backend())
         ffi.cdef("struct foo_s { int len; short data[5]; };")
@@ -1252,6 +1253,17 @@ class BackendTests:
         res = f(10, ffi.cast("int", 100), ffi.cast("long long", 1000))
         assert res == 20 + 300 + 5000
 
+    def test_callback_decorator(self):
+        ffi = FFI(backend=self.Backend())
+        #
+        @ffi.callback("long(long, long)", error=42)
+        def cb(a, b):
+            return a - b
+        #
+        assert cb(-100, -10) == -90
+        sz = ffi.sizeof("long")
+        assert cb((1 << (sz*8-1)) - 1, -10) == 42
+
     def test_unique_types(self):
         ffi1 = FFI(backend=self.Backend())
         ffi2 = FFI(backend=self.Backend())
@@ -1275,9 +1287,9 @@ class BackendTests:
         ffi = FFI(backend=self.Backend())
         ffi.cdef("typedef enum { Value0 = 0 } e, *pe;\n"
                  "typedef enum { Value1 = 1 } e1;")
-        assert ffi.getctype("e*") == 'enum $1 *'
-        assert ffi.getctype("pe") == 'enum $1 *'
-        assert ffi.getctype("e1*") == 'enum $2 *'
+        assert ffi.getctype("e*") == 'enum $e *'
+        assert ffi.getctype("pe") == 'enum $e *'
+        assert ffi.getctype("e1*") == 'enum $e1 *'
 
     def test_new_ctype(self):
         ffi = FFI(backend=self.Backend())
@@ -1293,7 +1305,6 @@ class BackendTests:
         assert ffi.string(e) == "AA"     # pick the first one arbitrarily
 
     def test_nested_anonymous_struct(self):
-        py.test.skip("later")
         ffi = FFI(backend=self.Backend())
         ffi.cdef("""
             struct foo_s {
@@ -1302,9 +1313,9 @@ class BackendTests:
             };
         """)
         assert ffi.sizeof("struct foo_s") == 3 * SIZE_OF_INT
-        p = ffi.new("struct foo_s *", [[1], [3]])
+        p = ffi.new("struct foo_s *", [1, 2, 3])
         assert p.a == 1
-        assert p.b == 0
+        assert p.b == 2
         assert p.c == 3
         assert p.d == 3
         p.d = 17
@@ -1314,6 +1325,47 @@ class BackendTests:
         assert p.b == 19
         assert p.c == 17
         assert p.d == 17
+        p = ffi.new("struct foo_s *", {'b': 12, 'd': 14})
+        assert p.a == 0
+        assert p.b == 12
+        assert p.c == 14
+        assert p.d == 14
+
+    def test_nested_anonymous_union(self):
+        ffi = FFI(backend=self.Backend())
+        ffi.cdef("""
+            union foo_u {
+                struct { int a, b; };
+                union { int c, d; };
+            };
+        """)
+        assert ffi.sizeof("union foo_u") == 2 * SIZE_OF_INT
+        p = ffi.new("union foo_u *", [5])
+        assert p.a == 5
+        assert p.b == 0
+        assert p.c == 5
+        assert p.d == 5
+        p.d = 17
+        assert p.c == 17
+        assert p.a == 17
+        p.b = 19
+        assert p.a == 17
+        assert p.b == 19
+        assert p.c == 17
+        assert p.d == 17
+        p = ffi.new("union foo_u *", {'d': 14})
+        assert p.a == 14
+        assert p.b == 0
+        assert p.c == 14
+        assert p.d == 14
+        p = ffi.new("union foo_u *", {'b': 12})
+        assert p.a == 0
+        assert p.b == 12
+        assert p.c == 0
+        assert p.d == 0
+        # we cannot specify several items in the dict, even though
+        # in theory in this particular case it would make sense
+        # to give both 'a' and 'b'
 
     def test_cast_to_array_type(self):
         ffi = FFI(backend=self.Backend())
@@ -1336,3 +1388,81 @@ class BackendTests:
         del q
         import gc; gc.collect(); gc.collect(); gc.collect()
         assert seen == [1]
+
+    def test_CData_CType(self):
+        ffi = FFI(backend=self.Backend())
+        assert isinstance(ffi.cast("int", 0), ffi.CData)
+        assert isinstance(ffi.new("int *"), ffi.CData)
+        assert not isinstance(ffi.typeof("int"), ffi.CData)
+        assert not isinstance(ffi.cast("int", 0), ffi.CType)
+        assert not isinstance(ffi.new("int *"), ffi.CType)
+
+    def test_CData_CType_2(self):
+        ffi = FFI(backend=self.Backend())
+        assert isinstance(ffi.typeof("int"), ffi.CType)
+
+    def test_bool(self):
+        ffi = FFI(backend=self.Backend())
+        assert int(ffi.cast("_Bool", 0.1)) == 1
+        assert int(ffi.cast("_Bool", -0.0)) == 0
+        assert int(ffi.cast("_Bool", b'\x02')) == 1
+        assert int(ffi.cast("_Bool", b'\x00')) == 0
+        assert int(ffi.cast("_Bool", b'\x80')) == 1
+        assert ffi.new("_Bool *", False)[0] == 0
+        assert ffi.new("_Bool *", 1)[0] == 1
+        py.test.raises(OverflowError, ffi.new, "_Bool *", 2)
+        py.test.raises(TypeError, ffi.string, ffi.cast("_Bool", 2))
+
+    def test_use_own_bool(self):
+        ffi = FFI(backend=self.Backend())
+        ffi.cdef("""typedef int bool;""")
+
+    def test_ordering_bug1(self):
+        ffi = FFI(backend=self.Backend())
+        ffi.cdef("""
+            struct foo_s {
+                struct bar_s *p;
+            };
+            struct bar_s {
+                struct foo_s foo;
+            };
+        """)
+        q = ffi.new("struct foo_s *")
+        bar = ffi.new("struct bar_s *")
+        q.p = bar
+        assert q.p.foo.p == ffi.NULL
+
+    def test_ordering_bug2(self):
+        ffi = FFI(backend=self.Backend())
+        ffi.cdef("""
+            struct bar_s;
+
+            struct foo_s {
+                void (*foo)(struct bar_s[]);
+            };
+
+            struct bar_s {
+                struct foo_s foo;
+            };
+        """)
+        q = ffi.new("struct foo_s *")
+
+    def test_addressof(self):
+        ffi = FFI(backend=self.Backend())
+        ffi.cdef("struct foo_s { int x, y; };")
+        p = ffi.new("struct foo_s *")
+        a = ffi.addressof(p[0])
+        assert repr(a).startswith("<cdata 'struct foo_s *' 0x")
+        py.test.raises(TypeError, ffi.addressof, p)
+        py.test.raises((AttributeError, TypeError), ffi.addressof, 5)
+
+    def test_addressof_field(self):
+        ffi = FFI(backend=self.Backend())
+        ffi.cdef("struct foo_s { int x, y; };")
+        p = ffi.new("struct foo_s *")
+        a = ffi.addressof(p[0], 'y')
+        assert repr(a).startswith("<cdata 'int *' 0x")
+        assert int(ffi.cast("uintptr_t", a)) == (
+            int(ffi.cast("uintptr_t", p)) + ffi.sizeof("int"))
+        assert a == ffi.addressof(p, 'y')
+        assert a != ffi.addressof(p, 'x')
