@@ -1,4 +1,4 @@
-import py, sys
+import py, sys, re
 from cffi import FFI, FFIError, CDefError, VerificationError
 
 class FakeBackend(object):
@@ -9,7 +9,7 @@ class FakeBackend(object):
     def sizeof(self, name):
         return 1
 
-    def load_library(self, name):
+    def load_library(self, name, flags):
         if sys.platform == 'win32':
             assert "msvcr" in name
         else:
@@ -43,6 +43,8 @@ class FakeBackend(object):
         return FakeType("<void>")
     def cast(self, x, y):
         return 'casted!'
+    def _get_types(self):
+        return "CData", "CType"
 
 class FakeType(object):
     def __init__(self, cdecl):
@@ -177,3 +179,23 @@ def test_override():
     assert C.foo.BType == '<func (), <int>, False>'
     ffi.cdef("long foo(void);", override=True)
     assert C.foo.BType == '<func (), <long>, False>'
+
+def test_cannot_have_only_variadic_part():
+    # this checks that we get a sensible error if we try "int foo(...);"
+    ffi = FFI()
+    e = py.test.raises(CDefError, ffi.cdef, "int foo(...);")
+    assert str(e.value) == \
+           "foo: a function with only '(...)' as argument is not correct C"
+
+def test_parse_error():
+    ffi = FFI()
+    e = py.test.raises(CDefError, ffi.cdef, " x y z ")
+    assert re.match(r'cannot parse " x y z "\n:\d+:', str(e.value))
+
+def test_cannot_declare_enum_later():
+    ffi = FFI()
+    e = py.test.raises(NotImplementedError, ffi.cdef,
+                       "typedef enum foo_e foo_t; enum foo_e { AA, BB };")
+    assert str(e.value) == (
+           "enum foo_e: the '{}' declaration should appear on the "
+           "first time the enum is mentioned, not later")
